"""oracle_attack.py

   This module is the python code that can be used to connect to the Oracle and decrypt the ciphertext. 
   It takes 3 arguments:

     1. -ip (--ipaddress):  The IP address of the machine to run the server.
     2. -p (--port):        The port to listen on.
     3. -id (--keyid):      The unique keyid that was assigned to each student       

   Feed 'oracle_attack.py' a 32 hex-byte chunk concatenated where:
   - The first 16 hex-bytes are the block (initialization vector) that can be used to manipulate the Oracle. 
   - The second 16 hex-bytes are the block of data that should be decrypted.

   Each student has a unique keyid assigned, which corresponds to a secret key that is unique to their encrypted text. 
   The server will respond to the client with 'Message decrypted successfully' (valid padding) if the input decrypts with successful padding 
   And "Padding error during decryption" (invalid padding) otherwise.
   The server will only check padding for 2 blocks at the time (iv + ciphertext), if more blocks are sent it will check padding on the last two blocks.

   *Note* A hex byte is two chars (00, fe, ab, etc..) so two blocks of size 16 should be 64 TOTAL in length!

   Example usage:
   python oracle_attack.py -ip shasta.cs.unm.edu -p 10008 -id 08

   Originally by Meisam Navaki, edited by Jed Crandall

"""
# Dependencies
import argparse
import socket
import sys
import binascii
from par import p

# Handle command-line arguments
parser = argparse.ArgumentParser()
parser.add_argument("-ip", "--ipaddress", help='ip address where the server is running', required=True)
parser.add_argument("-p", "--port", help='port where the server is listening on', required=True)
parser.add_argument("-id", "--keyid", help='unique key id', required=True)
args = parser.parse_args()

ipaddress = args.ipaddress
port = args.port
keyid = args.keyid.zfill(2)

cipher_hex = p[keyid] 

fn = "transcripts/transcript-" + str(keyid) + ".txt"
f = open(fn, "w")

# send the manipulated block to the server and get the result back
def is_valid(block):
    global ipaddress
    global port
    global keyid

    # Length of either response.
    # Server response may have spaces in it
    # to adhere to message length
    MESSAGE_LENGTH = 32
    BLOCK_SIZE = 32
    cipher_size = len(block)
    if not(cipher_size % BLOCK_SIZE == 0):
        print("Bad block(s) size")
        exit()

    else:
        # Create a TCP/IP socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # Connect the socket to the port where the server is listening
        server_address = (ipaddress, int(port))
        sock.connect(server_address)

        num_blocks = (cipher_size/BLOCK_SIZE)-1 # num_blocks to be decrypted (IV not counting as ciphertext block)

        # Send data
        message = block + ':' + str(int(num_blocks)) + ':' + keyid  
        f.write("C->S: " + message + '\n')
        sock.sendall(bytes(message, 'ascii'))

        # Look for the response
        amount_received = 0
        amount_expected = MESSAGE_LENGTH

        while amount_received < amount_expected:
            data = str(sock.recv(MESSAGE_LENGTH))
            amount_received += len(data)
        f.write("S->C: " + data + '\n')
        if data.find("error") != -1 or data.find("not valid") != -1:
            sock.close()
            return False
        else:
            sock.close()
            return True
    return False

def hex_to_list(iv):
    iv_list = []
    for i in range(int(len(iv)/2)):
        x = int(iv[i*2:(i+1)*2], 16)
        iv_list.append(x)
    return iv_list
  
def list_to_hex(iv):
    iv_hex = ''
    for byte in iv:
        iv_hex += "0x{:02x}".format(byte).split('x')[-1]
    return iv_hex

# Perform padding oracle attack for one block
def padding_oracle_attack_block(iv, c1, iv_o):

    iv = hex_to_list(iv)
    iv_o = hex_to_list(iv_o)
    plain = []
    for i in range(1,17):
        # for current test position, try all possible values
        for j in range(0,256):
            iv[-i] = j
            if is_valid(list_to_hex(iv) + c1):
                # the plaintext is our solution value, xor'd with
                # original iv, xor'd with current padding i.e. "i"
                # A byte decrypted
                plain.insert(0, j^iv_o[-i]^i)
                break
        for k in range(1, i + 1):
            iv[-k] = iv[-k] ^ i ^ (i + 1)
    return ''.join([chr(r) for r in plain])


# Perform padding oracle attack
def padding_oracle_attack(cipher):
    BLOCK_SIZE = 32
    num_blocks = int(len(cipher)/BLOCK_SIZE) - 2
    print('num_blocks is ' + str(num_blocks))
    plaintext = ''
    # iterate over all blocks and decrypt block by block
    # Note: the first block is IV
    for i in range(num_blocks):
        print("trying to decrypt block #" + str(i+1), "...")
        block_plain = padding_oracle_attack_block(cipher[i*32:(i+1)*32], cipher[(i+1)*32:(i+2)*32], cipher[i*32:(i+1)*32])
        print("block decrypted:", block_plain)
        plaintext += block_plain
    print("=====================")
    print("PLAINTEXT:", plaintext)




# Perform the oracle attack and decrypt the ciphertext
padding_oracle_attack(cipher_hex)

