"""Server.py
   This module is the 'padding oracle'. It takes 2 command-line arguments:

     1. -ip (--ipaddress): The IP address of the machine to run the server.
     2. -p (--port):       The port to listen on.
   
   server.py runs and listens on a port for a 32-byte chunk of data. The first 16-byte block
   is used as the initialization vector and the second 16-byte block is the encrypted data
   that is decrypted server-side and verified that proper padding is used. If proper padding
   is achieved, the server will respond with 'Valid' otherwise it will respond with 'Invalid'.
   This is all students need to work their way down and successfully decrypt the message.

   Example usage:
   python server.py -ip localhost -p 10000 
"""

# Dependencies
import argparse
import socket
import sys
from Crypto.Cipher import AES
from keys import keyids

#-------------------------------------------------------------------

# Size a key should adhere to in chars
KEY_SIZE = 16

# A block is 16 hex bytes, which will be 32 total chars, so a student should supply 64 chars
BLOCK_SIZE = 32

#-------------------------------------------------------------------

def find_key(key_id):
    if key_id in keyids:
        secret_key = keyids[key_id]
    return secret_key

#-------------------------------------------------------------------

def valid_keyid(key_id):
    valid_key = False

    if key_id in keyids:
        valid_key = True
    print(key_id + ' is the key ID')
    return valid_key

#-------------------------------------------------------------------

# Parse Command-Line Arguments
parser = argparse.ArgumentParser()
parser.add_argument("-ip", "--ipaddress")
parser.add_argument("-p", "--port")
args = parser.parse_args()

# Create TCP/IP socket
sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

# Bind socket to port
server_address = (args.ipaddress, int(args.port))
sys.stdout.write('Starting up on: %s port %s\n' % server_address)
sock.bind(server_address)

# Listen for incoming connections
sock.listen(10)

while True:
    sys.stdout.write('Waiting for a connection...\n') # Wait for a conneciton
    connection, client_address = sock.accept()
  
    try:
        sys.stdout.write('Connection from: ' + str(client_address) + '\n')

        # Receive the data in small chunks and retransmit it
        while True:
            message = connection.recv((BLOCK_SIZE * 11)+3+3).decode()
            print('Received: "%s"' % message)

            split_message = message.split(':') # receive and split the message where':'
            data = split_message[0] 
            key_id = ''.join(c for c in split_message[2] if c.isdigit())
            blocks_rcvd = split_message[1] 
            print('Blocks received', blocks_rcvd)

            valid_key = valid_keyid(key_id) 

            if valid_key == True:
                secret_key = find_key(key_id) # look for correspondent secret_key
                print('Key ID:', key_id)
                print('Secret Key: ', secret_key)
            else:
                sys.stdout.write('Key ID incorrect\n')
                # Connection sends 32 bytes regardless of valid or invalid that way the client knows when to stop listening. Hence, extra spaces.
                connection.sendall(bytes("Key ID is not valid. Try again  ", 'ascii'))
                break

            # Perform block split and check for valid padding
            #block_pair = get_2blocks(data) # block_pair = ciphertext split into 2 blocs
            iv = bytes.fromhex(data[0:32])
            cipher_text = bytes.fromhex(data[32:])
            aes = AES.new(secret_key.encode('ascii')[0:16], AES.MODE_CBC, iv)
            plain_text = aes.decrypt(cipher_text)
            print(plain_text)
            print(plain_text[-1])

            # If last byte is:         /x01 then it's valid
            # If last two bytes are:   /x02/x02 then it's valid
            # If last three bytes are: /x03/x03/x03 then it's valid
            # etc...
            valid_padding = False
            pt_len = plain_text[-1]
            if pt_len > 0 and pt_len <= 16:
                valid_padding = True
                for i in range(pt_len):
                    if plain_text[-1 - i] != pt_len:
                        valid_padding = False

            if valid_padding == True:
                sys.stdout.write('Sending VALID back to the client\n')
                # Connection sends 32 bytes regardless of valid or invalid that way the client knows when to stop listening. Hence, extra spaces.
                connection.sendall(bytes("Message decrypted successfully  ", 'ascii'))
                break

            elif valid_padding == False:
                sys.stdout.write('Sending INVALID back to the client\n')
                # Connection sends 32 bytes regardless of valid or invalid that way the client knows when to stop listening. Hence, extra spaces.
                connection.sendall(bytes("Padding error during decryption!", 'ascii'))
                break

            else:
                sys.stdout.write('No more data from: ' + client_address + '\n')
                break
  
    finally:
        # Clean up the connection
        connection.close()
